# -*- coding: utf-8 -*-
"""Return editions for a given ISBN."""

import logging

from ._core import EAN13
from .dev import vias
from ._exceptions import NotRecognizedServiceError, NotValidISBNError
from ._openled import query as oed
from ._thinged import query as ted

PROVIDERS = ('any', 'merge', 'openl', 'thingl')
TRUEPROVIDERS = ('openl', 'thingl')  # <-- by priority
LOGGER = logging.getLogger(__name__)


def fake_provider_any(isbn):
    """Fake provider 'any' service."""
    providers = {'openl': oed, 'thingl': ted}
    for provider in TRUEPROVIDERS:
        data = []
        try:
            data = providers[provider](isbn)
            if len(data) > 1:
                return data
            continue  # pragma: no cover
        except Exception:  # pragma: no cover
            continue
    return data  # pragma: no cover


def fake_provider_merge(isbn):
    """Fake provider 'merge' service."""
    try:  # pragma: no cover
        named_tasks = (('openl', oed), ('thingl', ted))
        results = vias.parallel(named_tasks, isbn)
        odata = results.get('openl', [])
        tdata = results.get('thingl', [])
        data = list(set(odata + tdata))
        return data
    except Exception:  # pragma: no cover
        return []


def editions(isbn, service='merge'):
    """Return the list of ISBNs of editions related with this ISBN."""
    isbn = EAN13(isbn)
    if not isbn:
        LOGGER.critical('%s is not a valid ISBN', isbn)
        raise NotValidISBNError(isbn)

    if service not in PROVIDERS:
        LOGGER.critical('%s is not a recognized editions provider', service)
        raise NotRecognizedServiceError(service)

    if service == 'merge':
        return fake_provider_merge(isbn)
    if service == 'any':
        return fake_provider_any(isbn)

    if service == 'openl':
        return oed(isbn)
    if service == 'thingl':
        return ted(isbn)



# -*- coding: utf-8 -*-
# flake8: noqa
# pylint: skip-file
"""
nose tests
"""

from nose.tools import assert_equals
from ..dev._bouth23 import u
from ..dev._fmt import _fmtbib

canonical = {
    'ISBN-13': u('9780123456789'),
    'Title': u('A book about nothing'),
    'Publisher': u('No Paper Press'),
    'Year': u('2000'),
    'Language': u('en'),
    'Authors': [u('John Smith'), u('José Silva')]
}


def test_fmtbib():
    """Test the formating into several bibliographic formats."""
    assert_equals(len(_fmtbib("bibtex", canonical)), 182)
    assert_equals(len(_fmtbib("labels", canonical)), 158)
    assert_equals(len(_fmtbib("endnote", canonical)), 103)
    assert_equals(len(_fmtbib("msword", canonical)), 485)
    assert_equals(len(_fmtbib("json", canonical)), 229)
    assert_equals(len(_fmtbib("csl", canonical)), 253)
    assert_equals(len(_fmtbib("refworks", canonical)), 130)
    assert_equals(len(_fmtbib("opf", canonical)), 861)



# -*- coding: utf-8 -*-
"""Use Google to get an ISBN from words from title and author's name."""

import logging

from ._core import get_canonical_isbn, get_isbnlike
from .dev import webservice

try:  # pragma: no cover
    from urllib.parse import quote
except ImportError:  # pragma: no cover

    def quote(x):  # pragma: no cover
        """Do nothing if PY2."""
        return x


LOGGER = logging.getLogger(__name__)


def goos(words):
    """Use Google to get an ISBN from words from title and author's name."""
    service_url = "http://www.google.com/search?q=ISBN+"
    search_url = service_url + quote(words.replace(' ', '+'))

    user_agent = 'w3m/0.5.3'

    content = webservice.query(
        search_url,
        user_agent=user_agent,
        appheaders={
            'Content-Type': 'text/plain; charset="UTF-8"',
            'Content-Transfer-Encoding': 'Quoted-Printable',
        })
    isbns = get_isbnlike(content)

    for item in isbns:
        isbn = get_canonical_isbn(item, output='isbn13')
        if isbn:
            break
    if not isbns or not isbn:  # pragma: no cover
        LOGGER.debug('No ISBN found for %s', words)
        return None
    return isbn



# -*- coding: utf-8 -*-
"""Query web services."""

import gzip
import logging

from ._bouth23 import bstream, s
from ._exceptions import ISBNLibHTTPError, ISBNLibURLError
from .. import config

# pylint: disable=import-error
# pylint: disable=wrong-import-order
# pylint: disable=no-name-in-module
try:  # pragma: no cover
    from urllib.parse import urlencode
    from urllib.request import Request, urlopen
    from urllib.error import HTTPError, URLError
except ImportError:  # pragma: no cover
    from urllib import urlencode
    from urllib2 import Request, urlopen, HTTPError, URLError

UA = 'isbnlib (gzip)'
LOGGER = logging.getLogger(__name__)


# pylint: disable=too-few-public-methods
class WEBService(object):
    """Class to query web services."""

    def __init__(self, url, user_agent=UA, values=None, appheaders=None):
        """Initialize main properties."""
        self._url = url
        # headers to accept gzipped content
        headers = {'Accept-Encoding': 'gzip', 'User-Agent': user_agent}
        # add more user provided headers
        if appheaders:  # pragma: no cover
            headers.update(appheaders)
        # if 'data' it does a POST request (data must be urlencoded)
        data = urlencode(values).encode('utf8') if values else None
        self._request = Request(url, data, headers=headers)

    def response(self):
        """Check errors on response."""
        try:
            response = urlopen(self._request, timeout=config.URLOPEN_TIMEOUT)
            LOGGER.debug('Request headers:\n%s', self._request.header_items())
        except HTTPError as e:  # pragma: no cover
            LOGGER.critical('ISBNLibHTTPError for %s with code %s [%s]',
                            self._url, e.code, e.msg)
            if e.code in (401, 403, 429):
                raise ISBNLibHTTPError(
                    '%s Are you making many requests?' % e.code)
            if e.code in (502, 504):
                raise ISBNLibHTTPError(
                    '%s Service temporarily unavailable!' % e.code)
            raise ISBNLibHTTPError('(%s) %s' % (e.code, e.msg))
        except URLError as e:  # pragma: no cover
            LOGGER.critical('ISBNLibURLError for %s with reason %s', self._url,
                            e.reason)
            raise ISBNLibURLError(e.reason)
        return response if response else None

    def data(self):
        """Return the uncompressed data."""
        res = self.response()
        LOGGER.debug('Response headers:\n%s', res.info())
        if res.info().get('Content-Encoding') == 'gzip':
            buf = bstream(res.read())
            f = gzip.GzipFile(fileobj=buf)
            data = f.read()
        else:  # pragma: no cover
            data = res.read()
        return s(data)


def query(url, user_agent=UA, values=None, appheaders=None):
    """Query to a web service."""
    service = WEBService(
        url, user_agent=user_agent, values=values, appheaders=appheaders)
    data = service.data()
    LOGGER.debug('Raw data from service:\n%s', data)
    return data



# -*- coding: utf-8 -*-
# flake8: noqa
# pylint: skip-file
"""
nose tests
"""

from ..dev._helpers import last_first, cutoff_tokens, parse_placeholders, fake_isbn
from nose.tools import assert_equals


def test_last_first():
    """Test the parsing of author's name into (Surname, First Name)."""
    assert_equals(last_first("Surname, First Name"),
                  {"last": "Surname",
                   "first": "First Name"})
    assert_equals(last_first("First Name Surname"),
                  {"last": "Surname",
                   "first": "First Name"})
    assert_equals(last_first("Surname1, First1 and Sur2, First2"),
                  {"last": "Surname1",
                   "first": "First1 and Sur2, First2"})


def test_cutoff_tokens():
    """Test the 'cutoff_tokens' function."""
    assert_equals(cutoff_tokens(['1', '23', '456'], 3), ['1', '23'])


def test_parse_placeholders():
    """Test the parsing of placeholders."""
    assert_equals(parse_placeholders('{isbn}_akaj_{name}'), ['{isbn}', '{name}'
                                                             ])


def test_fake_isbn():
    """Test the 'fake_isbn' function."""
    assert_equals(fake_isbn(' Hello?? Wer,  ! ksDf:  asdf. ; '),
                  '1111006407537')
    assert_equals(fake_isbn(' Hello?? Wer,  ! ksDf:  asdf. ; ',
                            author=''),
                  '1108449680873')
    assert_equals(fake_isbn(' Hello?? Wer,  ! ksDf:  asdf. ; ',
                            author=' '),
                  '1108449680873')
    assert_equals(fake_isbn(' Hello?? Wer,  ! ksDf:  asdf. ; ',
                            author='',
                            publisher=''),
                  '1181593982422')
    assert_equals(fake_isbn(' Hello?? Wer,  ! ksDf:  asdf. ; ',
                            author='a',
                            publisher='K'),
                  '1895031085488')
    assert_equals(fake_isbn(' Hello?? Wer,  ! ksDf:  asdf. ; ',
                            author='A',
                            publisher='k'),
                  '1895031085488')



# -*- coding: utf-8 -*-
"""Config file for isbnlib."""

# --> Import only external modules! <--

# Timeouts
URLOPEN_TIMEOUT = 10  # seconds
THREADS_TIMEOUT = 12  # seconds


# URLOPEN_TIMEOUT is used by webservice
def seturlopentimeout(seconds):  # pragma: no cover
    """Set the value of URLOPEN_TIMEOUT (in seconds)."""
    global URLOPEN_TIMEOUT
    URLOPEN_TIMEOUT = seconds


# THREADS_TIMEOUT is a parameter used downstream by Thread calls (see vias.py)
def setthreadstimeout(seconds):  # pragma: no cover
    """Set the value of THREADS_TIMEOUT (in seconds)."""
    global THREADS_TIMEOUT
    THREADS_TIMEOUT = seconds


# API keys
apikeys = {}


def add_apikey(service, apikey):  # pragma: no cover
    """Add API keys."""
    apikeys[service] = apikey


# Generic Options
options = {'VIAS_MERGE': 'parallel'}


def set_option(option, value):  # pragma: no cover
    """Set the value for option."""
    options[option] = value



# -*- coding: utf-8 -*-
"""Query the Open Library for related ISBNs."""

import logging

from . import get_canonical_isbn, get_isbnlike
from .dev._bouth23 import u
from .dev.webquery import query as wquery

LOGGER = logging.getLogger(__name__)
UA = 'isbnlib (gzip)'
SERVICE_URL = 'http://openlibrary.org/query.json?type=/type/edition&'\
              '{selectors}'
CODES = 'isbn_13={isbn}&books='
ISBNS = '{code}&isbn_13='


def query(isbn):
    """Query the Open Library for related ISBNs."""
    try:
        data = wquery(
            SERVICE_URL.format(selectors=CODES.format(isbn=isbn)),
            user_agent=UA)
        codes = [rec['key'] for rec in data]
        isbnlikes = [isbn]
        for code in codes:
            txt = wquery(
                SERVICE_URL.format(selectors=ISBNS.format(code=code)),
                user_agent=UA,
                parser=None)
            isbnlikes.extend(get_isbnlike(txt))
        isbns = [get_canonical_isbn(isbnlike) for isbnlike in isbnlikes]
        isbns = sorted(list(set([u(n) for n in isbns if n])))  # qa: py26
    except Exception:  # pragma: no cover, qa: FIXME
        LOGGER.debug('No data from Open Library for isbn %s', isbn)
    return isbns



# -*- coding: utf-8 -*-
# flake8: noqa
# pylint: skip-file
""" nose tests

"""

from nose.tools import assert_equals, assert_raises
from ..dev import Metadata, stdmeta
from ..dev._bouth23 import u


def test_stdmeta():
    """Test the transformation of raw records into standard metadata."""
    # test stdmeta from data
    r = {
        'ISBN-13': u('9780123456789 '),
        'Title': u('Bla. Bla /Title .'),
        'Publisher': u(''),
        'Year': u('2000'),
        'Language': u('en'),
        'Authors': [u('author1. mba'), u('author2   ')]
    }
    R = {
        'ISBN-13': u('9780123456789'),
        'Title': u('Bla. Bla /Title'),
        'Publisher': u(''),
        'Year': u('2000'),
        'Language': u('en'),
        'Authors': [u('author1. mba'), u('author2')]
    }
    A = {
        'ISBN-13': u('9780123456789 '),
        'Title': b'Bla. Bla /Title .',
        'Publisher': u(''),
        'Year': b'2000',
        'Language': u('en'),
        'Authors': [u('author1. mba'), u('author2   ')]
    }
    B = {
        'ISBN-13': u('9780123456789'),
        'Title': u('Bla. Bla /Title .'),
        'Publisher': u(''),
        'Year': u('2000'),
        'Language': u('en'),
        'Authors': u('author1')
    }
    assert_equals(stdmeta(r), R)
    assert_equals(stdmeta(R), R)
    assert_raises(Exception, stdmeta, A)
    assert_raises(Exception, stdmeta, B)


def test_metaclass():
    """Test the creation of a Metadata class from raw records."""
    R = {
        'ISBN-13': u('9780123456789'),
        'Title': u('Bla. Bla /Title'),
        'Publisher': u(''),
        'Year': u('2000'),
        'Language': u('en'),
        'Authors': [u('author1. mba'), u('author2')]
    }
    dt = Metadata(R)
    assert_equals(dt.value, R)



# -*- coding: utf-8 -*-
# flake8: noqa
# pylint: skip-file

from nose.tools import assert_equals, raises

from .._exceptions import NotRecognizedServiceError, NotValidISBNError
from .._ext import editions

# nose tests



def test_editions_openl():
    """Test the 'openl editions' service."""
    assert_equals(len(editions('9780099536017', service='openl')) > 4, True)


def test_editions_thingl():
    """Test the 'thingl editions' service."""
    assert_equals(len(editions('9780151446476', service='thingl')) > 19, True)


def test_editions_any():
    """Test the 'any editions' service."""
    assert_equals(len(editions('9780151446476', service='any')) > 19, True)


def test_editions_merge():
    """Test the 'merge editions' service."""
    assert_equals(len(editions('9780151446476', service='merge')) > 19, True)


@raises(NotValidISBNError)
def test_editions_NotValidISBNError():
    """Test the 'editions' service error detection (NotValidISBNError)."""
    editions('978')


@raises(NotRecognizedServiceError)
def test_editions_NotRecognizedServiceError():
    """Test the 'editions' service error detection (NotRecognizedServiceError)."""
    editions('9780156001311', service='xxx')



# -*- coding: utf-8 -*-
# flake8: noqa
# isort:skip_file
"""Library to validate, clean, transform and get metadata of ISBN strings (for devs)."""

# Define isbnlib API and set lib environment

import logging as _logging

from ._exceptions import (quiet_errors, ISBNLibException,
                          NotRecognizedServiceError, NotValidISBNError,
                          PluginNotLoadedError)

# configuration
from . import config  # <-- first import

# main modules
from ._core import (is_isbn10, is_isbn13, to_isbn10, to_isbn13, check_digit10,
                    check_digit13, clean, canonical, notisbn, get_isbnlike,
                    get_canonical_isbn, GTIN13, EAN13, RE_ISBN10, RE_ISBN13,
                    RE_LOOSE, RE_NORMAL, RE_STRICT)
from ._ext import (cover, desc, mask, meta, info, editions, isbn_from_words,
                   doi, ren)
from ._goom import query as goom
from ._doitotex import doi2tex

# Ranges Database date
from ._data.data4mask import RDDATE

# config _logging for lib
_nh = _logging.NullHandler()
_logging.getLogger('isbnlib').addHandler(_nh)

# alias
ISBN13 = EAN13
ean13 = EAN13

# dunders

__all__ = ('is_isbn10', 'is_isbn13', 'clean', 'check_digit10', 'check_digit13',
           'mask', 'info', 'meta', 'to_isbn10', 'to_isbn13', 'get_isbnlike',
           'notisbn', 'ean13', 'EAN13', 'cover', 'desc', 'canonical',
           'get_canonical_isbn', 'editions', 'isbn_from_words', 'quiet_errors',
           'config', '__version__', '__support__', 'doi', 'ren', 'ISBN13',
           'GTIN13', 'ISBNLibException', 'NotRecognizedServiceError',
           'NotValidISBNError', 'PluginNotLoadedError', 'goom', 'doi2tex',
           'RDDATE')

__version__ = '3.9.3'  # <-- literal IDs
__support__ = 'py27, py34, py35, py36, py37, pypy, pypy3'  # <-- literal IDs



# -*- coding: utf-8 -*-
"""Return metadata, of a given DOI, formated as BibTeX."""

import logging

from .dev.webservice import query

LOGGER = logging.getLogger(__name__)

URL = 'http://dx.doi.org/{DOI}'
UA = 'isbnlib (gzip)'


def doi2tex(doi):
    """Get the bibtex ref for doi."""
    data = query(
        URL.format(DOI=doi),
        user_agent=UA,
        appheaders={
            'Accept': 'application/x-bibtex; charset=utf-8',
        })
    if not data:  # pragma: no cover
        LOGGER.warning('no data return for doi: %s', doi)
        return None
    return data



# -*- coding: utf-8 -*-
# flake8: noqa
# pylint: skip-file
"""
nose tests
"""

from ..dev.webservice import query as wsquery
from nose.tools import assert_equals


def test_webservice():
    """Test that values can be passed to a WebService query."""
    assert_equals(len(repr(wsquery("http://example.org", values={'some': 'values'}))) > 0,
                  True)



# -*- coding: utf-8 -*-
from . import helpers, vias
from ._data import Metadata, stdmeta
from ._exceptions import (
    DataNotFoundAtServiceError, DataWrongShapeError, ISBNLibDevException,
    ISBNLibHTTPError, ISBNLibURLError, NoAPIKeyError, NoDataForSelectorError,
    NotValidMetadataError, RecordMappingError, ServiceIsDownError)
from .webquery import WEBQuery
from .webservice import WEBService

# alias
ISBNToolsDevException = ISBNLibDevException
ISBNToolsHTTPError = ISBNLibHTTPError
ISBNToolsURLError = ISBNLibURLError

__all__ = ('ISBNToolsDevException', 'ISBNLibDevException',
           'ISBNToolsHTTPError', 'ISBNToolsURLError', 'ISBNLibHTTPError',
           'ISBNLibURLError', 'DataNotFoundAtServiceError',
           'ServiceIsDownError', 'DataWrongShapeError',
           'NotValidMetadataError', 'NoDataForSelectorError',
           'RecordMappingError', 'NoAPIKeyError', 'Metadata', 'stdmeta',
           'WEBService', 'WEBQuery', 'vias', 'helpers')



# -*- coding: utf-8 -*-
"""Get image links of the book's cover."""

import logging

from .dev.webquery import query as wquery

LOGGER = logging.getLogger(__name__)

UA = "isbnlib (gzip)"
SERVICE_URL = 'https://www.googleapis.com/books/v1/volumes?q=isbn+{isbn}'\
              '&fields=items/volumeInfo(imageLinks)&maxResults=1'


def cover(isbn):
    """Main entry point for cover."""
    from .registry import metadata_cache  # <-- dynamic
    # check the cache first
    cache = metadata_cache
    if cache:  # pragma: no cover
        key = 'img-url-go-' + isbn
        try:  # pragma: no cover
            if cache[key]:
                return cache[key]
            else:
                raise KeyError  # <-- IMPORTANT: caches don't return error!
        except KeyError:  # pragma: no cover
            pass
    # request to the web service
    data = wquery(SERVICE_URL.format(isbn=isbn), user_agent=UA)
    try:
        lnks = data['items'][0]['volumeInfo']['imageLinks']
        # put in cache
        if cache and lnks:  # pragma: no cover
            cache[key] = lnks
        return lnks
    except (KeyError, IndexError):  # pragma: no cover
        LOGGER.debug('No cover img data for %s', isbn)
    return None



# -*- coding: utf-8 -*-
"""Registry for metadata services, formatters and cache."""

from pkg_resources import iter_entry_points

from . import _goob as goob
from . import _openl as openl
from ._imcache import IMCache
from .dev._fmt import _fmtbib

# SERVICES

services = {
    'default': goob.query,
    'goob': goob.query,
    'openl': openl.query,
}


def setdefaultservice(name):  # pragma: no cover
    """Set the default service."""
    global services
    services['default'] = services[name]


def add_service(name, query):  # pragma: no cover
    """Add a new service to services."""
    global services
    services[name] = query


# FORMATTERS

bibformatters = {
    'default': lambda x: _fmtbib('labels', x),
    'labels': lambda x: _fmtbib('labels', x),
    'bibtex': lambda x: _fmtbib('bibtex', x),
    'csl': lambda x: _fmtbib('csl', x),
    'json': lambda x: _fmtbib('json', x),
    'opf': lambda x: _fmtbib('opf', x),
    'endnote': lambda x: _fmtbib('endnote', x),
    'refworks': lambda x: _fmtbib('refworks', x),
    'msword': lambda x: _fmtbib('msword', x),
}  # pragma: no cover


def setdefaultbibformatter(name):  # pragma: no cover
    """Set the default formatter."""
    global bibformatters
    bibformatters['default'] = bibformatters[name]


def add_bibformatter(name, formatter):  # pragma: no cover
    """Add a new formatter to formatters."""
    global bibformatters
    bibformatters[name] = formatter


def load_plugins():  # pragma: no cover
    """Load plugins with groups: isbnlib.metadata & isbnlib.formatters."""
    # get metadata plugins from entry_points
    try:
        for entry in iter_entry_points(group='isbnlib.metadata'):
            add_service(entry.name, entry.load())
    except:
        pass
    # get formatters from entry_points
    try:
        for entry in iter_entry_points(group='isbnlib.formatters'):
            add_bibformatter(entry.name, entry.load())
    except:
        pass


# load plugins on import
load_plugins()

# CACHE
metadata_cache = IMCache()  # should be an instance


def set_cache(cache):  # pragma: no cover
    """Set cache for metadata."""
    global metadata_cache
    metadata_cache = cache


custom_cache = None  # should be an instance


def set_custom_cache(cache):  # pragma: no cover
    """Set a 'spare' cache."""
    global custom_cache
    custom_cache = cache
